// --- [ Servo_main.c ] ------------------------------------------------
//
//     tab = 3 -  24.09.2012
//
// Dient zur Ansteuerung eines Modellbau-Servos ber die TWI-Schnittstelle,
// Aufruf:
//           TWI-Adresse, Code, Parameter entsprechend gewnschter Funktion
// Code [1   .. 3 ], Wert       - Offset fr jeweilen Servo setzen
// Code [10 ... 30], Wert       - Ruheposition fr jeweiligen Servo setzen
// Code [11 ... 31], Wert       - Drehgeschwindigkeit fr jeweiligen Servo setzen
// Code [255], Wert, Wert, Wert - Offset fr alle Servos nacheinander setzen
// Code [131]                   - Alle Neutralpositionen ins Eeprom schreiben
// Code [133]                   - Alle Offsets ins Eeprom schreiben
// Code [135]                   - Alle Speed-Werte ins Eeprom schreiben

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <util/delay.h>
#include "USI_Slave.h"
#include "eeprom.h"
#include "global.h"

#define SERVOS             3              // Anzahl der Servos, das sind am ATTiny25 max 3 Stck
                                          // die Pins, an denen die Servos angeschlossen sind
#define SERVOS_OFF()       (PORTB = PORTB & (~(1<<PB1 | 1<<PB3 | 1<<PB4)))
#define SERVO0_ON()        (PORTB = PORTB | (1<<PB1))
#define SERVO1_ON()        (PORTB = PORTB | (1<<PB3))
#define SERVO2_ON()        (PORTB = PORTB | (1<<PB4))

#define USI_DATA_LEN       8              // die Gre des Array, mit dem Daten aus dem USI-Modul ausgetauscht werden

#define NEUTRAL_POS        187            // theoretisch 1.5ms = 1500us / 8us = 187
#define ABWEICHUNG         40             // wie weit darf die Neutralposition verschoben werden
#define NEUTRAL_MIN        (NEUTRAL_POS - ABWEICHUNG)
#define NEUTRAL_MAX        (NEUTRAL_POS + ABWEICHUNG)

#define OCR1A_MIN          80             // Zur Sicherheit die extremen OCR1A-Wert definieren, normalerweise:
#define OCR1A_MAX          300            // - Minimal 0.9ms =  900us / 8us = 112
                                          // - Neutral 1.5ms = 1500us / 8us = 187
                                          // - Maximal 2.1ms = 2100us / 8us = 263 (> uint8_t !)
#define SPEED_MAX          10             // hchster Wert fr die Drehgeschwindigkeit
#define SPEED_DEFAULT      0              // Vorgabe, wenn es keinen im Eeprom abgelegten Wert gibt

#define AUSSCHLAG          75             // Maximal zulssige Ablenkung aus der Neutralposition, normalerweise
                                          // - Ausschlag +-0.6ms = 600us / 8us = 75
                                          // Soll verhindern, dass der Servo in den Anschlag geht

#define EEPROM_ADR         16             // ab hier liegen die Daten fr die Ruheposition, Offset und speed im EEPROM

void Make_OCR1A(uint8_t idx);
void USI_Error(uint8_t rx);

// --- Globale Variablen ------------------------------------------------------

uint8_t  speed[SERVOS];                   // Drehgeschwindigkeit der Servos (0 = Max, 1 (Langsam) bis 10 (Schnell))
int8_t   offset[SERVOS];                  // Offset der Zielposition bezogen auf den Neutralpunkt
uint8_t  neutral_pos[SERVOS];             // Ruheposition des Servos als OCR1A-Wert
uint16_t target_pos[SERVOS];              // Servo-Ziel-Position als OCR1A-Wert
uint16_t last_pos[SERVOS];                // Letzte Servo-Position in OCR1A-Einheiten
                                          // wird bentigt bei langsamer Servo-Bewegung
uint8_t usi_data[USI_DATA_LEN];           // damit werden die Daten mit dem USI-Modul ausgetauscht

                                          // die Register GPIOR0, GPIOR1 und GPIOR2 werden benutzt !

// ----------------------------------------------------------------------------
int main( void )
{
   uint8_t rx, i, j;
   uint16_t servo_pos = 0;

   CLKPR = (1<<CLKPCE);
   CLKPR = (0<<CLKPS3 | 0<<CLKPS2 | 1<<CLKPS1 | 1<<CLKPS0); // Prescale 8 -> 1 MHz cPU-Takt

   ACSR = (1<<ACD);                          // Comparator disabled
   PRR  = (0<<PRTIM1 | 0<<PRTIM0 | 0<<PRUSI | 1<<PRADC);
                                             // ungenutzte Module abschalten
   USI_SLA_Init();

   asm volatile("sei");	 							// enable_interrupt

   DDRB   |= (1<< PB1 | 1<<PB3 | 1<<PB4);    // PB1/3/4 als Ausgang

// der Timer 0 setzt den 20ms Takt
   TCCR0A = (1<<WGM01 | 0<<WGM00);           // CTC-Mode
   TCCR0B = (0<<WGM02 | 1<<CS02 | 0<<CS01 | 0 << CS00); // Prescale = 256, Takt = 128us -> OCR0A = 78
   OCR0A  = 78;
   TIMSK  = (1<<OCIE0A);

// der Timer 1 ist fr die Impulslnge zustndig
//   TCCR1 = (1<<CTC1 | 0<<CS13 | 1<<CS12 | 0<<CS11 | 0<<CS10); // Prescale 8, 8us

// einige Parameter aus dem Eeprom laden - sofern dort etwas gespeichert ist
// ist dort 0xFF eingetragen, dann den Default-Wert whlen
   i = EEPROM_ADR;                           // ab dieser Eeprom-Adresse sind die Daten abgelegt

   for (j = 0; j < SERVOS; j++)
   {
                                             // zurerst wird die Neutralposition gelesen und geprft
      neutral_pos[j] = EEPROM_Read(i++);
      if (neutral_pos[j] > NEUTRAL_MAX) neutral_pos[j] = NEUTRAL_POS;
      else if (neutral_pos[j] < NEUTRAL_MIN) neutral_pos[j] = NEUTRAL_POS;

                                             // dahinter folgt im EEprom der Offset von der Neutralposition
      offset[j] = (int8_t) EEPROM_Read(i++);
      if (offset[j] > AUSSCHLAG) offset[j] = 0;
      else if (offset[j] < (0 - AUSSCHLAG)) offset[j] = 0;

      target_pos[j] = neutral_pos[j] + offset[j];  // alle Servos starten in Ruheposition + Offset
      last_pos[j] = target_pos[j];

      speed[j] = EEPROM_Read(i++);           // die Drehgeschindigkeit aller Servos einlesen
      if (speed[j] > SPEED_MAX) speed[j] = SPEED_DEFAULT;

   }
   EEARH = 0;                                // Adresse auf Eeprom-Anfang zurcksetzen
   EEARL = 0;

   while(1)
   {
      if ((rx = USI_SLA_Rx_Cnt()))           // sind Daten im Puffer ?
      {                                      // ist eine bertragung im Gange, wird hier bis zum Abschluss gewartet
         USI_SLA_Get_Data(usi_data); 			// Daten aus dem Empfangspuffer nach usi_data[] holen

         if (rx == 3)                        // Die Adresse, der Code sowie genau ein Parameter wurden bergeben
                                             // -> Servo + Offset aus Ruheposition
                                             // -> Servoruheposition
         {
            if (usi_data[0] == 0)            // Einen General Call ignorieren
            {
               USI_Error(rx);                // Fehlercode ins USI-Modul schreiben
               continue;
            }

            j = usi_data[1];
            if (j && (j <= SERVOS))           // 1 / 2 / 3 ... -> Offset fr einzelnen Servo wird bergeben
            {
               j--;                           // der Index ist immer um 1 niedriger als die Servo-Nr.
               offset[j] = (int) usi_data[2];
               Make_OCR1A(j);
            }

            else if (((j % 10) == 0) && j)    // 10 / 20 / 30 ...
            {
               j /= 10;                       // den Index ermitteln
               j--;
               neutral_pos[j] = usi_data[2];
               if (neutral_pos[j] < NEUTRAL_MIN) neutral_pos[j] = NEUTRAL_MIN;
               else if (neutral_pos[j] > NEUTRAL_MAX) neutral_pos[j] = NEUTRAL_MAX;
               Make_OCR1A(j);
            }

            else if (((j % 10) == 1) && j)    // 11 / 21 / 31 ...
            {
               j--;                           // zuerst den Einer abziehen
               j /= 10;                       // den Index ermitteln
               j--;
               if (usi_data[2] <= SPEED_MAX)
               {
                  speed[j] = usi_data[2];
               }

               else
               {
                  USI_Error(rx);
               }

            }

            else                             // keine interpretierbaren Daten gefunden
            {
               USI_Error(rx);
               continue;                     // Nichts tun, wenn falsche Codes bermittelt wurden
            }

                                             // die Daten des aktuellen Servos ber TWI bereitstellen
            usi_data[0] = (j + 1);                       // Servo Nr.
            usi_data[1] = neutral_pos[j];                //
            usi_data[2] = offset[j];
            usi_data[3] = target_pos[j] / 256;           // high-Byte
            usi_data[4] = (uint8_t) target_pos[j];       // low-Byte
            usi_data[5] = speed[j];
            for (i = 6; i < USI_DATA_LEN; i++) usi_data[i] = 0;

            USI_SLA_Put_Data(usi_data, USI_DATA_LEN);   // usi_data ins USI-Modul exportieren, kann dann ausgelesen werden
         }

         else if (rx == (SERVOS + 2))                    // TWI-Adresse, Code plus 1 Byte je Servo
         {
            if (usi_data[1] == 255)                      // Code  "255"
            {
               for (i = 0; i < SERVOS; i++)
               {
                  offset[i] = usi_data[i+2];             // die Positionswerte bernehmen
                  Make_OCR1A(i);
                  usi_data[i] = target_pos[i];           // zum Debuggen vorbereiten
               }

            for (i = SERVOS; i < USI_DATA_LEN; i++) usi_data[i] = 0; // mit Nullen auffllen
            USI_SLA_Put_Data(usi_data, USI_DATA_LEN);    // zum Debuggen in den Ausgangspuffer schreiben
            }
         }

         else if (rx == 2)
         {
            if (usi_data[1] == 131)                   // Code "131" schreibt Neutralposition ins Eeprom
            {
               i = EEPROM_ADR;
               for (j = 0; j < SERVOS; j++)
               {
                  EEPROM_Write(i, neutral_pos[j]);
                  i += 3;                             // die Adresse wird jeweils um 2 incrementiert
                  usi_data[j] = neutral_pos[j];       // die Wert auch fr den USI_TX_Buffer vorbereiten
               }
             }

            else if (usi_data[1] == 133)              // Code "133" schreibt die Offset-Werte ins Eeprom
            {
               i = EEPROM_ADR + 1;                    // der Offset wird jeweils hinter die Neutralposition geschrieben
               for (j = 0; j < SERVOS; j++)
               {
                  EEPROM_Write(i, offset[j]);
                  i += 3;                             // die Adresse wird jeweils um 2 incrementiert
                  usi_data[j] = offset[j];            // die Wert auch fr den USI_TX_Buffer vorbereiten
               }
            }

            else if (usi_data[1] == 135)              // Code "135" schreibt die Speed-Werte ins Eeprom
            {
               i = EEPROM_ADR + 2;                    // peed wird jeweils hinter den Offset geschrieben
               for (j = 0; j < SERVOS; j++)
               {
                  EEPROM_Write(i, speed[j]);
                  i += 3;                             // die Adresse wird jeweils um 3 incrementiert
                  usi_data[j] = speed[j];             // die Wert auch fr den USI_TX_Buffer vorbereiten
               }

            }
            else
            {
               USI_Error(rx);
               continue;
            }

            for (i = SERVOS; i < USI_DATA_LEN; i++) usi_data[i] = 0; // mit Nullen auffllen
            USI_SLA_Put_Data(usi_data, USI_DATA_LEN);   // zum Debuggen in den Ausgangspuffer schreiben

            EEARH = 0;
            EEARL = 0;
         }
         else                                            // keine interpretierbaren Daten gefunden
         {
            USI_Error(rx);
         }
      }
                                                         // Hier werden die Impulse an die Servos geschickt
      if (GPIOR0)                                        // jeweils im 20ms Takt direkt nacheinander an alle Servos
      {
         GPIOR0 = FALSE;
         GPIOR2++;                                       // wird zur "Entschleunigung" des Servos genutzt

         for(i = 0; i < SERVOS; i++)
         {
            if      (i == 0) SERVO0_ON();                // den zustndigen Pin auf high setzen
            else if (i == 1) SERVO1_ON();
            else if (i == 2) SERVO2_ON();

 // Falls eine Verzgerung in speed[] definiert ist, hier den bergang herstellen
 // speed[] == 0 ignoriert Verzgerungen und legt direkt den neuen Wert an OCR1A an
 // wenn speed[] > 0, dann wird in der Schrittweite von speed an die Zielposition angenhert
 // GPIOR2 bildet dabei einen Teiler, der nur nach jedem 2. Durchgang die OCR1A Werte ndert
            if (speed[i])
            {
 //            if ((GPIOR2 % 3) == 0)                    // so wird nur bei jedem dritten Impuls incrementiert
               if (GPIOR2 & 0x01)                        // bei ungradzahligem GPIOR2 nicht tun
               {                                         // fhrt zur Verlangsamung der Drehung um Faktor 2
                  servo_pos = last_pos[i];               // nicht tun, OCR1A bleibt unverndert
               }

               else if (last_pos[i] < target_pos[i])      // Vernderung zu hheren Werten
               {
                  if ((last_pos[i] + speed[i]) < target_pos[i])
                  {
                     servo_pos = last_pos[i] + speed[i];
                  }

                  else
                  {
                     servo_pos = target_pos[i];
                  }
               }

               else if (last_pos[i] > target_pos[i])     // Vernderung zu kleineren Werten verndern
               {
                  if ((last_pos[i] - speed[i]) > target_pos[i])
                  {
                     servo_pos = last_pos[i] - speed[i];
                  }

                  else
                  {
                     servo_pos = target_pos[i];
                  }
               }

               else                                      // last_pos == target_pos
               {
                  servo_pos = target_pos[i];
               }

            }

            else                                         // if speed[i] == 0
            {
               servo_pos = target_pos[i];                // direkt die Zielposition anfahren
            }

            last_pos[i] = servo_pos;                     // fr jeden Servo die aktuelle Position merken

// Falls der OCR1A-Wert nicht in das 8Bit-Register passt, dann die Impulslnge in zwei Schritten erzeugen:
// zuerst mit OCR1A = 255 beginnen und
// danach noch einen Rundgang mit dem Rest anschlieen
// Das Register GPIOR1 nimmt den Wert fr die Dauer der Verlngerung im zweiten Rundgang auf
            if (servo_pos < 256)
            {
               OCR1A = servo_pos;
               GPIOR1 = 0;
            }

            else                                         // wenn servo_pos > uint8_t
            {
               OCR1A  = 255;                             // erste Wartezeit mit Wert 255
               GPIOR1 = servo_pos - 255;                 // die verbleibende Wartezeit zwischenspeichern
            }

            TIMSK = (0<<OCIE0A | 1<<OCIE1A);             // Interrupt enablen
            TCNT1 =  0;                                  // den Timer/Counter1 starten
            TCCR1 = (1<<CTC1 | 0<<CS13 | 1<<CS12 | 0<<CS11 | 0<<CS10);

            while(TCCR1);                                 // warten bis die ISR TTCR1 auf 0 setzt (den Timer abschaltet)
         }
      }
   }
}


// ------------------------------------------------------------------------
// Berechnet aus der Ruheposition sowie dem Offset den erforderlichen OCR1A-Wert (target_pos[])
void Make_OCR1A(uint8_t idx)
{
   // den maximalen Ausschlag aus der Ruheposition begrenzen -> +- AUSSCHLAG
   if (offset[idx] > AUSSCHLAG) offset[idx] = AUSSCHLAG;
   else if (offset[idx] < (0 - AUSSCHLAG)) offset[idx] = (0 - AUSSCHLAG);

   // den maximalen Ausschlag des Servos begrenzen -> OCR1A_MIN bis OCR1A_MAX
   target_pos[idx] = neutral_pos[idx] + offset[idx];

   if (target_pos[idx] > OCR1A_MAX) target_pos[idx] = OCR1A_MAX;
   else if (target_pos[idx] < OCR1A_MIN) target_pos[idx] = OCR1A_MIN;
}

// ---------------------------------------------------------------------------
// Schreibt die empfangenen Daten in den Ausgangspuffer des USI-Moduls,
// und fllt mit 255 auf -> als Fehlerhinweis
// Wenn ausfhrbare Daten empfangen wurden, dann ist der Puffer mit 0 aufgefllt !

void USI_Error(uint8_t rx)
{
   uint8_t i;
   for (i = rx; i < USI_DATA_LEN; i++) usi_data[i] = 255;
   USI_SLA_Put_Data(usi_data, USI_DATA_LEN);   // usi_data ins USI-Modul exportieren, kann dann ausgelesen werden
}

#if ASM == FALSE                                // Die Compare_ISR's knnen als C oder Assembler-Code eingebunden werden
// -----------------------------
// Wird alle 20ms aufgerufen und startet eine Impulsfolge
// als Flag wird das Register GPIOR0 benutzt
ISR(TIMER0_COMPA_vect)
{
   GPIOR0 = 1;
}

// -------------------------------
// wird am Ende des Impulses aufgerufen und legt den Pin wieder auf low
// hngt ggf. noch eine Verlngerung an, wenn die Dauer grer als 255 betragen hat
ISR(TIMER1_COMPA_vect)
{
   if (GPIOR1)                                     // hier steht der Wert fr die Verlngerung drin
   {
      OCR1A  = GPIOR1;
      GPIOR1 = 0;
   }

   else
   {
      SERVOS_OFF();                                // alle Pins auf low legen
      TCCR1 = 0;
      TIMSK = (1<<OCIE0A | 0<<OCIE1A);             // Interrrupt disablen
   }
}
#endif
// --- [ eof ] ----------------------------------------------------------------

